/* eslint-disable no-console */
import React, { useEffect, useState } from 'react';
import { Dialog } from '@jupyterlab/apputils';

import { StylesProvider } from '@material-ui/core';

import { URLExt } from '@jupyterlab/coreutils';
import { ServerConnection, ContentsManager } from '@jupyterlab/services';
import { ISettingRegistry } from '@jupyterlab/settingregistry';
import { JobsView, Scheduler } from '@jupyterlab/scheduler';
import { CommandRegistry } from '@lumino/commands';

import { CreateNotebookJobForm } from './CreateNotebookJobForm';
import { errorCodes, i18nStrings, JUPYTER_COMMAND_IDS } from '../constants';
import { renderTextWithLinks } from '../utils';
import * as Styles from './styles';

import { parseListSageMakerImagesAPIResponse } from '../utils';
import { ImagesMap } from '../types';
import { getJupyterLabTheme } from '../themeProvider';
import { ThemeProvider } from '@mui/material/styles';

type CreateNotebookJobProps = Scheduler.IAdvancedOptionsProps & {
  requestClient: ServerConnection.ISettings;
  contentsManager: ContentsManager;
  commands: CommandRegistry;
  settingRegistry: ISettingRegistry;
};

const showCredentialsErrorDialog = async (commands: CommandRegistry) => {
  const body = (
    <>
      {i18nStrings.Dialog.awsCredentialsError.body.text.map((text, idx) => (
        <p key={idx} className={Styles.CredentialsErrorDialogParagraph}>
          {renderTextWithLinks(text, i18nStrings.Dialog.awsCredentialsError.body.links)}
        </p>
      ))}
    </>
  );
  const dialog = new Dialog({
    title: i18nStrings.Dialog.awsCredentialsError.title,
    body,
    buttons: [
      Dialog.cancelButton(),
      Dialog.okButton({ label: i18nStrings.Dialog.awsCredentialsError.buttons.enterKeysInTerminal }),
    ],
  });
  const result = await dialog.launch();
  if (result.button.label === i18nStrings.Dialog.awsCredentialsError.buttons.enterKeysInTerminal) {
    commands.execute(JUPYTER_COMMAND_IDS.terminal.createNew);
  }
};

const CreateNotebookJob: React.FC<CreateNotebookJobProps> = ({
  requestClient,
  contentsManager,
  commands,
  jobsView,
  errors,
  handleErrorsChange,
  ...rest
}) => {
  const fetchExecutionEnvironments = async () => {
    const url = URLExt.join(requestClient.baseUrl, '/advanced_environments');
    const response = await ServerConnection.makeRequest(url, {}, requestClient);

    if (response.status !== 200) {
      const responseJson = await response.json();
      const responseErrorCode = responseJson.error_code;
      const awsCredentialsErrors = Object.values(errorCodes.awsCredentials);
      if (awsCredentialsErrors.indexOf(responseErrorCode) >= 0) {
        showCredentialsErrorDialog(commands);
      }
      return null;
    } else {
      const executionEnvironmentsResponse = await response.json();
      return executionEnvironmentsResponse;
    }
  };

  const fetchKernelAndImages = async () => {
    const url = URLExt.join(requestClient.baseUrl, '/sagemaker_images');
    const response = await ServerConnection.makeRequest(url, {}, requestClient);

    if (response.status !== 200) {
      return {};
    } else {
      return parseListSageMakerImagesAPIResponse(await response.json());
    }
  };

  useEffect(() => {
    const loadingErrors = {
      ...errors,
      environmentsStillLoading: 'EnvironmentsStillLoadingError',
      kernelsStillLoading: 'KernelsStillLoadingError',
    };

    handleErrorsChange(loadingErrors);
    const { model } = rest;

    if (jobsView === JobsView.CreateForm) {
      fetchExecutionEnvironments()
        .then(async (executionEnvironments: any) => {
          setEnvironmentsLoading(false);
          setExecutionEnvironments(executionEnvironments);
        })
        .catch((error) => console.error(error));
    }

    fetchKernelAndImages()
      .then(async (imagesMap: ImagesMap) => {
        setImagesMap(imagesMap);
        setKernelsLoading(false);

        if (jobsView === JobsView.JobDetail || jobsView === JobsView.JobDefinitionDetail) {
          if (model.runtimeEnvironmentParameters?.sm_image) {
            setPreSelectedNotebookKernelSpec(model.runtimeEnvironmentParameters?.sm_image as string);

            // set other loading states to false since we are not making these async actions anymore
            setEnvironmentsLoading(false);
            loadingErrors.kernelsStillLoading = '';
            return;
          }
        }
      })
      .catch((error) => console.error(error));
    // eslint-disable-next-line react-hooks/exhaustive-deps
  }, [jobsView, rest.model.inputFile]);

  const [executionEnvironments, setExecutionEnvironments] = useState<any>({});
  const [imagesMap, setImagesMap] = useState<any>();
  const [preSelectedNotebookKernelSpec, setPreSelectedNotebookKernelSpec] = useState<string>();

  // loading states for async operations
  const [kernelsLoading, setKernelsLoading] = useState<boolean>(true);
  const [environmentsLoading, setEnvironmentsLoading] = useState<boolean>(true);

  const loading = kernelsLoading || environmentsLoading;

  if (loading) {
    return null;
  }

  if (jobsView === JobsView.CreateForm && !executionEnvironments?.auto_detected_config) {
    return null;
  }

  return (
    <ThemeProvider theme={getJupyterLabTheme()}>
      <StylesProvider injectFirst>
        <CreateNotebookJobForm
          executionEnvironments={executionEnvironments}
          preSelectedNotebookKernelSpec={preSelectedNotebookKernelSpec}
          imagesMap={imagesMap}
          requestClient={requestClient}
          jobsView={jobsView as any}
          errors={errors}
          handleErrorsChange={handleErrorsChange}
          {...rest}
        />
      </StylesProvider>
    </ThemeProvider>
  );
};

export { CreateNotebookJob };
