import React, { SyntheticEvent, useState } from 'react';
import { i18nStrings } from '../../../constants';
import { Scheduler } from '@jupyterlab/scheduler';
import { Checkbox } from '@material-ui/core';
import InfoIcon from '@mui/icons-material/Info';
import { AutocompleteChangeReason } from '@mui/material/Autocomplete';
import { InputContainer } from '../InputContainer';
import { VpcCheckbox } from './VpcCheckbox/VpcCheckbox';
import { FormState } from '../../CreateNotebookJobForm';
import { EnvironmentVariables, IEnvironmentVariable } from './EnvironmentVariables';
import { MultiSelectContainer } from '../MultiSelectContainer';
import { Link, LinkTarget } from '../../../components/link';
import { Tooltip } from '../../../components/tooltip/Tooltip';
import { ServerConnection } from '@jupyterlab/services';
import { URLExt } from '@jupyterlab/coreutils';
import { DropdownItem } from '../../../components/selectinput';

import {
  validateRoleArn,
  validateS3Url,
  validateSecurityGroups,
  validateSubnets,
  validateKMS,
  validateImage,
  validateKernel,
} from './validationHelpers';

import * as Styles from '../../styles';

const widgetStrings = i18nStrings.ScheduleNoteBook.MainPanel.AdvancedOptions;
const tooltipsStrings = i18nStrings.ScheduleNoteBook.MainPanel.Tooltips;
const errorStrings = i18nStrings.ScheduleNoteBook.MainPanel.ErrorMessages;

const KMSLink = 'https://docs.aws.amazon.com/sagemaker/latest/dg/create-notebook-auto-execution-advanced.html';

interface AdvancedOptionProps {
  isDisabled: boolean;
  formState: FormState;
  setFormState: (state: FormState) => void;
  formErrors: Scheduler.ErrorsType;
  lccOptions: string[];
  environmentVariables: IEnvironmentVariable[];
  setEnvironmentVariables: (environmentParameters: IEnvironmentVariable[]) => void;
  availableSecurityGroups: string[];
  availableSubnets: string[];
  initialSubnets: string[];
  requestClient: ServerConnection.ISettings;
  initialSecurityGroups: string[];
  userDefaultValues: FormState;
  setSubnets: (subnets: string[]) => void;
  setRoleArn: (roleArn: string) => void;
  setSecurityGroups: (securityGroups: string[]) => void;
  onSelectLCCScript: (item: DropdownItem | string) => void;
  handleChange: (e: React.ChangeEvent<HTMLInputElement | HTMLTextAreaElement>) => void;
  setFormValidationErrors: (errors: Scheduler.ErrorsType) => void;
  isVPCDomain: boolean;
  enableVPCSetting: boolean;
  setEnableVPCSetting: (checkBoxState: boolean) => void;
}

const kmsTooltipArea = (
  <div>
    <span className={Styles.TooltipTextContainer}>{tooltipsStrings.kmsKeyTooltip}</span>
    <Link href={KMSLink} target={LinkTarget.External}>
      <p className={Styles.TooltipLink}>{tooltipsStrings.LearnMore}</p>
    </Link>
  </div>
);

const AdvancedOptions: React.FunctionComponent<AdvancedOptionProps> = ({
  isDisabled,
  formState,
  formErrors,
  environmentVariables,
  setEnvironmentVariables,
  lccOptions,
  availableSecurityGroups,
  availableSubnets,
  initialSubnets,
  initialSecurityGroups,
  isVPCDomain,
  requestClient,
  enableVPCSetting,
  userDefaultValues,
  setFormState,
  handleChange,
  setSubnets,
  setSecurityGroups,
  onSelectLCCScript,
  setFormValidationErrors,
  setEnableVPCSetting,
  setRoleArn,
}) => {
  const [enableJobEncryption, setEnableJobEncryption] = useState(false);
  const [loading, setLoading] = useState(false);

  const validateVolumeFilePath = async (filePath: string) => {
    const url = URLExt.join(requestClient.baseUrl, '/validate_volume_path');
    setLoading(true);
    const response = await ServerConnection.makeRequest(
      url,
      { method: 'POST', body: JSON.stringify({ file_path: filePath }) },
      requestClient,
    );
    setLoading(false);

    if (response.status === 200) {
      const responseBody = await response.json();
      if (responseBody.file_path_exist !== true) {
        setFormValidationErrors({
          ...formErrors,
          efsFilePathError: errorStrings.AdvancedOptions.EFSFilePathError,
        });
        return;
      }
    }

    setFormValidationErrors({
      ...formErrors,
      efsFilePathError: '',
    });
  };

  const disableAddEnvironmentVariablesButton = () =>
    isDisabled || environmentVariables.length >= 48 || !!formErrors.environmentVariablesError;

  const validateInitScript = (e: React.ChangeEvent<HTMLInputElement | HTMLTextAreaElement>) => {
    const filePath = e.target.value;
    const isAllWhiteSpace = filePath.trim().length === 0;

    if (!isAllWhiteSpace) {
      validateVolumeFilePath(filePath);
    } else {
      setFormValidationErrors({
        ...formErrors,
        efsFilePathError: '',
      });
    }
  };

  const handleImageOnBlur = (e: React.ChangeEvent<HTMLInputElement | HTMLTextAreaElement>) => {
    const { value } = e.target;
    const errorMessage = validateImage(value);

    setFormValidationErrors({
      ...formErrors,
      ImageError: errorMessage,
    });
  }

  const handleKernelOnBlur = (e: React.ChangeEvent<HTMLInputElement | HTMLTextAreaElement>) => {
    const { value } = e.target;
    const errorMessage = validateKernel(value);

    setFormValidationErrors({
      ...formErrors,
      KernelError: errorMessage,
    });
  }

  const handleS3InputFieldOnBlur = (e: React.ChangeEvent<HTMLInputElement | HTMLTextAreaElement>) => {
    const name = e.target.name;
    const errorMessage = validateS3Url(e.target.value);

    setFormValidationErrors({
      ...formErrors,
      [name === 's3_input' ? 's3InputFolderError' : 's3OutputFolderError']: errorMessage,
    });
  };

  const handleKMSOnBlur = (e: React.ChangeEvent<HTMLInputElement | HTMLTextAreaElement>) => {
    const name = e.target.name;
    const errorMessage = validateKMS(e.target.value);
    setFormValidationErrors({
      ...formErrors,
      [name === 'sm_output_kms_key' ? 'outputKMSError' : 'ebsKMSError']: errorMessage,
    });
  };

  const handleRoleArnOnBlur = (e: React.ChangeEvent<HTMLInputElement | HTMLTextAreaElement>) => {
    const { value } = e.target;
    const errorMessage = validateRoleArn(value);
    setRoleArn(value);

    setFormValidationErrors({
      ...formErrors,
      roleError: errorMessage,
    });
  };

  const handleChangeSecurityGroups = (event: SyntheticEvent, values: string[], reason: AutocompleteChangeReason) => {
    const [errorMessageSecurityGroup, errorMessageSubnet] = validateSecurityGroups(values, formState.vpc_subnets);

    setSecurityGroups(values);

    setFormValidationErrors({
      ...formErrors,
      securityGroupError: errorMessageSecurityGroup ?? '',
      subnetError: errorMessageSubnet ?? formErrors.subnetError,
    });
  };

  const handleChangeSubnets = (event: SyntheticEvent, values: string[], reason: AutocompleteChangeReason) => {
    const [errorMessageSubnet, errorMessageSecurityGroup] = validateSubnets(values, formState.vpc_security_group_ids);

    setSubnets(values);

    setFormValidationErrors({
      ...formErrors,
      securityGroupError: errorMessageSecurityGroup ?? formErrors.securityGroupError,
      subnetError: errorMessageSubnet ?? '',
    });
  };

  return (
    <div className={Styles.WidgetFieldsContainer}>
      <InputContainer
        name={'sm_image'}
        onChange={handleChange}
        disabled={loading}
        readOnly={isDisabled}
        required
        value={formState.sm_image}
        placeholder={widgetStrings.Placeholders.ImagePlaceHolder}
        labelInfo={widgetStrings.Image}
        errorMessage={formErrors.ImageError}
        onBlur={handleImageOnBlur}
        toolTipText={tooltipsStrings.ImageTooltipText}
      />

      <InputContainer
        name={'sm_kernel'}
        onChange={handleChange}
        disabled={loading}
        readOnly={isDisabled}
        required
        value={formState.sm_kernel}
        placeholder={widgetStrings.Placeholders.KernelPlaceHolder}
        labelInfo={widgetStrings.Kernel}
        errorMessage={formErrors.KernelError}
        onBlur={handleKernelOnBlur}
        toolTipText={tooltipsStrings.KernelTooltipText}
      />

      <InputContainer
        name={'role_arn'}
        disabled={loading}
        readOnly={isDisabled}
        required
        labelInfo={widgetStrings.RoleArn}
        errorMessage={formErrors.roleError}
        placeholder={widgetStrings.Placeholders.RolePlaceHolder}
        onChange={handleChange}
        value={formState.role_arn}
        onBlur={handleRoleArnOnBlur}
        toolTipText={tooltipsStrings.RoleArnTooltip}
      />

      <InputContainer
        name={'s3_input'}
        onChange={handleChange}
        required
        disabled={loading}
        readOnly={isDisabled}
        value={formState.s3_input}
        placeholder={widgetStrings.Placeholders.S3BucketPlaceHolder}
        labelInfo={widgetStrings.s3InputFolder}
        errorMessage={formErrors.s3InputFolderError}
        onBlur={handleS3InputFieldOnBlur}
        toolTipText={tooltipsStrings.InputFolderTooltip}
      />
      <InputContainer
        name={'s3_output'}
        onChange={handleChange}
        required
        disabled={loading}
        readOnly={isDisabled}
        value={formState.s3_output}
        placeholder={widgetStrings.Placeholders.S3BucketPlaceHolder}
        labelInfo={widgetStrings.s3OutputFolder}
        errorMessage={formErrors.s3OutputFolderError}
        onBlur={handleS3InputFieldOnBlur}
        toolTipText={tooltipsStrings.OutputFolderTooltip}
      />

      {!isDisabled && (
        <div className={Styles.TooltipCheckBoxContainer}>
          <Checkbox
            name={'kms_checkbox'}
            className={Styles.Checkbox}
            color={'primary'}
            checked={enableJobEncryption}
            onChange={(e) => {
              const checkedState = e.target.checked;
              setEnableJobEncryption(checkedState);
              // when checkbox gets checked, the user wants to enable the kms encryption, so
              // we should pick the kms key from the user default values object and check it for validity.
              // if they disable the kms encryption we should clear out the kms key and the respective form errors.
              const outputKMSKey = checkedState ? userDefaultValues.sm_output_kms_key : '';
              const volumeKMSKey = checkedState ? userDefaultValues.sm_volume_kms_key : '';
              setFormState({
                ...formState,
                sm_output_kms_key: outputKMSKey,
                sm_volume_kms_key: volumeKMSKey,
              });
              setFormValidationErrors({
                ...formErrors,
                outputKMSError: validateKMS(outputKMSKey),
                ebsKMSError: validateKMS(volumeKMSKey),
              });
            }}
          />
          <label>{widgetStrings.enableEncryption}</label>
          <Tooltip
            classes={{
              popperInteractive: Styles.PopperInteractive,
            }}
            title={kmsTooltipArea}
            interactive
          >
            <InfoIcon fontSize="small" />
          </Tooltip>
        </div>
      )}

      {enableJobEncryption && (
        <>
          <InputContainer
            name={'sm_output_kms_key'}
            onChange={handleChange}
            required={false}
            readOnly={isDisabled}
            disabled={loading}
            value={formState.sm_output_kms_key}
            placeholder={isDisabled ? widgetStrings.Placeholders.NoneSelected : widgetStrings.enterKMSArnOrID}
            labelInfo={widgetStrings.kmsKey}
            errorMessage={formErrors.outputKMSError}
            onBlur={handleKMSOnBlur}
            toolTipText={!isDisabled ? tooltipsStrings.kmsKeyTooltip : undefined}
          />
          <InputContainer
            name={'sm_volume_kms_key'}
            onChange={handleChange}
            required={false}
            readOnly={isDisabled}
            disabled={loading}
            value={formState.sm_volume_kms_key}
            placeholder={isDisabled ? widgetStrings.Placeholders.NoneSelected : widgetStrings.enterKMSArnOrID}
            labelInfo={widgetStrings.ebsKey}
            errorMessage={formErrors.ebsKMSError}
            onBlur={handleKMSOnBlur}
            toolTipText={!isDisabled ? tooltipsStrings.ebsKeyTooltip : undefined}
          />
        </>
      )}

      {!isDisabled && (
        <VpcCheckbox
          isChecked={enableVPCSetting}
          setChecked={setEnableVPCSetting}
          initialSecurityGroups={initialSecurityGroups}
          initialSubnets={initialSubnets}
          availableSubnets={availableSubnets}
          formState={formState}
          formErrors={formErrors}
          setFormErrors={setFormValidationErrors}
          setFormState={setFormState}
        />
      )}

      {(enableVPCSetting || isDisabled) && (
        <div>
          <MultiSelectContainer
            required
            name={'vpc_subnets'}
            disabled={isDisabled || availableSubnets.length === 0}
            label={widgetStrings.subnet}
            options={availableSubnets}
            value={formState.vpc_subnets}
            onChange={handleChangeSubnets}
            errorMessage={formErrors.subnetError}
            placeholder={`${widgetStrings.Placeholders.SelectPrivateSubnets}`}
            tooltip={tooltipsStrings.SubnetTooltip}
            disabledTooltip={`${widgetStrings.Placeholders.NoPrivateSubnets}`}
            freeSolo
          />

          <MultiSelectContainer
            required
            className={'securityGroupSelector'}
            name={'vpc_security_group_ids'}
            disabled={isDisabled || availableSubnets.length === 0}
            label={widgetStrings.securityGroup}
            options={availableSecurityGroups}
            value={formState.vpc_security_group_ids}
            onChange={handleChangeSecurityGroups}
            errorMessage={formErrors.securityGroupError}
            placeholder={`${widgetStrings.Placeholders.selectOrAdd} ${widgetStrings.securityGroup}`}
            tooltip={tooltipsStrings.SecurityGroupsTooltip}
            disabledTooltip={`${widgetStrings.Placeholders.No} ${widgetStrings.securityGroup}`}
            freeSolo
          />
        </div>
      )}

      <EnvironmentVariables
        isButtonDisabled={disableAddEnvironmentVariablesButton()}
        allFieldsDisabled={isDisabled}
        environmentVariables={environmentVariables}
        setEnvironmentVariables={setEnvironmentVariables}
        formErrors={formErrors}
        setFormErrors={setFormValidationErrors}
      />
      <div>
        <InputContainer
          placeholder={isDisabled ? widgetStrings.Placeholders.NoneSelected : widgetStrings.efsPlaceholder}
          labelInfo={widgetStrings.efsLabel}
          required={false}
          value={formState.sm_init_script}
          name={'sm_init_script'}
          readOnly={isDisabled}
          disabled={loading}
          errorMessage={formErrors.efsFilePathError}
          onChange={handleChange}
          onBlur={validateInitScript}
          toolTipText={!isDisabled ? tooltipsStrings.InitialScriptTooltip : undefined}
        />
        {/* {loading && <ActivityIndicator />} */}
      </div>
    </div>
  );
};

export default AdvancedOptions;
